#include <ucontext.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <jack/jack.h>
#include <pcl.h>

#include "rollendurchmesserzeitsammler.h"


// For DATASTART and DATAEND
#include <gc/gc.h>

// Using DATASTART and DATAEND
#include "gcconfig.h"


static coroutine_t dsp_coroutine;
static float *block;
static jack_port_t *port;
static jack_client_t *client;

static void *stack_top;
static void *stack_bot;

static tar_heap_t *heap;


static int get_blocksize(void){
  return jack_get_buffer_size(client);
}

static void out_sample(int i,float val){
  //printf("%d: val: %f\n",i,val);
  block[i]=val;
}


static void schemedebug(int val){
  //fprintf(stderr,"sd: %d\n",val);
}

//static void schemedebug2(void *val){
//  fprintf(stderr,"sd2: %p\n",val);
//}

int scheme_main(void);
static void dsp_coroutine_func(void *arg){
  int dummy;
  stack_top=&dummy;
  scheme_main();
}

ucontext_t ctxt;

static void block_ready(void){
  int dummy;

  // This is not necessary. Remove if everything works, and scratch head if it stops working after being removed.
#if 0
  if (getcontext(&ctxt) < 0){
    fprintf(stderr,"Getcontext failed.\n");
    exit(-5);
  }
#endif
  //tar_add_root(heap,&ctxt,((char*)&ctxt)+sizeof(ucontext_t));

  stack_bot=&dummy;
  //printf("Stack bot: %p, top: %p. size: %dk\n",stack_bot,stack_top,(((char*)stack_top)-((char*)stack_bot))/1024);
  co_resume();
}

static float between(float start,float end){
  return start + (end-start)*((float)random()/(float)RAND_MAX);
}

static int intbetween(int start,int end){
  return start + (end-start)*((float)random()/(float)RAND_MAX);
}


#define main scheme_main
#  include "graincloud.c"
#undef main

//int stack[32*4096];

static int process (jack_nframes_t nframes, void *arg)
{

  //fprintf(stderr,"entering\n");
  if(tar_entering_audio_thread(heap)==false){
    fprintf(stderr,"Using too much cpu. Skipping.\n");
    return 0;
  }
  //fprintf(stderr,"got it\n");

  block = jack_port_get_buffer (port, nframes);
  co_call(dsp_coroutine);
  //fprintf(stderr,"Stack bot: %p, top: %p\n",stack_bot,stack_top);

  if(tar_leave_audio_thread(heap)){

    // Add stack
    tar_add_root(heap,stack_bot,stack_top);
    //tar_add_root(heap,&stack[0],&stack[32*4096+1]);
    
    // Add registers
    tar_add_root(heap,dsp_coroutine,((char*)dsp_coroutine)+sizeof(ucontext_t));

    // Add data
    tar_add_root(heap,((char*)DATASTART),((char*)DATAEND));

    //    printf("Schemedata size:  %p %p %d\n",DATASTART,DATAEND,((char*)DATAEND)-((char*)DATASTART));
    //printf("Schemedata stack: %p %p %d\n",&stack[0],&stack[32*4096+1],0);

    // Signal gc to run.
    tar_run_gc(heap);
  }
  //fprintf(stderr,"exiting\n");
  return 0;
}



static void print_error(FILE *where,char *fmt, ...) {
  char temp[10000];
  va_list ap;
  va_start(ap, fmt);{
    vsnprintf (temp, 9998, fmt, ap);
  }va_end(ap);
  //syslog(LOG_INFO,temp);
  fprintf(where,"tar snapshot benchmark: %s\n",temp);
}

static int set_pid_priority(pid_t pid,int policy,int priority,char *message,char *name){
  struct sched_param par={0};
  par.sched_priority=priority;
  if((sched_setscheduler(pid,policy,&par)!=0)){
    print_error(stderr,message,pid,name,strerror(errno));
    return 0;
  }
  return 1;
}

void set_realtime(void){
  set_pid_priority(0,SCHED_FIFO,60,"Unable to set SCHED_FIFO for %d (\"%s\"). (%s)", "the xmessage fork");
}

void* threadstart_realtime(void *arg){
  tar_threadstart threadstart=(tar_threadstart)arg;
  set_realtime();
  threadstart();
  return NULL;

}

void create_realtime_thread(tar_threadstart threadstart){
  pthread_t *thread;
  thread=calloc(sizeof(pthread_t*),1);
  pthread_create(thread,NULL,threadstart_realtime,threadstart);
}

void* threadstart_mark(void *arg){
  tar_threadstart threadstart=(tar_threadstart)arg;
  //set_pid_priority(0,SCHED_FIFO,0,"Unable to set SCHED_FIFO for %d (\"%s\"). (%s)", "the xmessage fork");
  threadstart();
  return NULL;
}

void create_mark_thread(tar_threadstart threadstart){
  pthread_t *thread;
  thread=calloc(sizeof(pthread_t*),1);
  pthread_create(thread,NULL,threadstart_mark,threadstart);
}


int main(int argc,char **argv){

  // Set the values accessed by DATASTART and DATAEND
  GC_init();

  dsp_coroutine=co_create(dsp_coroutine_func,NULL,NULL,32*4096);
  //fprintf(stderr,"ai: %p\n",dsp_coroutine);

  fprintf(stderr,"1\n");
  // Init gc
  tar_init(1024*1024*10, // Static heap. (size doesn't matter for performance)
	    0.2*1024*1024, // Non-static heap. (the larger the size, the more cpu use)
	    4*1024*1024,
	    create_realtime_thread,
	    create_mark_thread);
  heap=tar_new_heap();
  fprintf(stderr,"2\n");
  // init jack.
  client = jack_client_new ("schemetest");

  if(client==NULL){
    fprintf(stderr,"Could not connect to jack\n");
    exit(1);
  }
  port = jack_port_register (client, "out", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

  // Init scheme code.
  fprintf(stderr,"ai-1\n");
  co_call(dsp_coroutine);
  fprintf(stderr,"ai0\n");
#if 0
  int i;
  for(i=0;i<4;i++){
    fprintf(stderr,"ai1\n");
    co_call(dsp_coroutine);
    fprintf(stderr,"ai2\n");
  }
  exit(0);
#endif

  fprintf(stderr,"3\n");

  // Start process.
  jack_set_process_callback (client, process, NULL);
  jack_activate (client);
  jack_connect (client, jack_port_name(port),"system:playback_1");
  jack_connect (client, jack_port_name(port),"system:playback_2");

  fprintf(stderr,"4\n");
  {
    char in[1000];
    gets(in);//sleep(10);
  }

#if 0
  int i;
  for(i=0;i<4;i++){
    co_call(dsp_coroutine);
  }
#endif

  return 0;
}
